#ifndef _SCREEN_TRANSITION_H
#define _SCREEN_TRANSITION_H

#include "types.h"

// extern ? D_8032FF60;
// extern ? D_8032FF64;
// extern ? D_8032FF68;

// extern ? func_802CAAE0(?);
// extern ? func_802CAB60(?);
// extern ? func_802CADB4(?);
// extern ? func_802CAF38(?);
// extern ? func_802CB0E4(?);
// extern ? func_802CB140(?);
// extern ? func_802CB19C(?);
// extern ? func_802CB274(?);
// extern ? func_802CB384(?);
// extern ? func_802CB400(?);
// extern ? func_802CB484(?);
// extern ? func_802CB508(?);
// extern ? func_802CB6A0(?);
// extern ? func_802CB9F8(?);
extern u32 func_802CC108();
// extern ? func_802CC2E8(?);
// extern ? Geo18_802CD1E8(?);

#endif /* _SCREEN_TRANSITION_H */
